#!/usr/bin/perl

use DBI;
use Carp qw(cluck croak);
use Data::Dumper;
#use lib qw( /usr/local/rrdtool-1.2.11/lib/perl );
use RRDs;
use Time::localtime;

use strict;

my $dbh = DBI->connect("dbi:mysql:cacti_plugins", 'cactiuser', 'cactiuser') || die "Could not connect: " . $DBI::errstr;

my $consolidation_functions = { 1 => 'AVERAGE',
                                2 => 'MIN',
                                3 => 'MAX',
                                4 => 'LAST',
                                AVERAGE => 1,
                                MIN     => 2,
                                MAX     => 3,
                                LAST    => 4 };

my $cdef_operators = { 1 => '+',
                       2 => '-',
                       3 => '*',
                       4 => '/',
                       5 => '%'};

my $cdef_functions = {1 => "SIN",
                      2 => "COS",
                      3 => "LOG",
                      4 => "EXP",
                      5 => "FLOOR",
                      6 => "CEIL",
                      7 => "LT",
                      8 => "LE",
                      9 => "GT",
                      10 => "GE",
                      11 => "EQ",
                      12 => "IF",
                      13 => "MIN",
                      14 => "MAX",
                      15 => "LIMIT",
                      16 => "DUP",
                      17 => "EXC",
                      18 => "POP",
                      19 => "UN",
                      20 => "UNKN",
                      21 => "PREV",
                      22 => "INF",
                      23 => "NEGINF",
                      24 => "NOW",
                      25 => "TIME",
                      26 => "LTIME" };

# Depending on the type of cdef in the database, it is dereferenced in different ways:
my $cdef_item_types = { 1 => $cdef_functions,   # Use the value as a key (Function)
                        2 => $cdef_operators,   # Use the value as a key (Operator)
                        4 => undef,             # Use the native value (Special Data Source)
                        5 => \&get_cdef_by_id,  # Recurse (Another CDEF)
                        6 => undef              # Use the native value (Custom String)
                      };

my $rra_id = shift @ARGV;

my $go_to_start = 0;  # If set, work the timings from the start of the day, week, month or year
if (defined ($ARGV[0]) && $ARGV[0] eq '--start')
{
    $go_to_start = 1;
}

unless (defined ($rra_id) && $rra_id ne "")
{
    print <<"EOF";
Usage: $0 [RRA ID] [ --start ]

For each RRA you have set up in cacti, your should set up a cron-job to run the reports
for the time-period that the rra represents. On standard cacti installations, the RRA
ID's given below for the daily/weekly/monthly and yearly reports should work:

# Daily Report
0 0 * * * perl /data/cacti/scripts/runReports.pl 1 >/dev/null 2>&1
# Weekly Reports
0 0 * * 1 perl /data/cacti/scripts/runReports.pl 2 >/dev/null 2>&1
# Monthly Reports
0 0 1 * * perl /data/cacti/scripts/runReports.pl 3 >/dev/null 2>&1
# Yearly Reports
0 0 1 1 * perl /data/cacti/scripts/runReports.pl 4 >/dev/null 2>&1

If you're running the reports from the command-line, you can specify '--start' which will
back-date the reports to the begining of the day, week, month or year.

EOF
    exit(1);
}

my @reportDefinitions;
if ($rra_id == 0)
{
    @reportDefinitions = getRows("SELECT
                                    plugin_cactireport_definitions.id AS report_id,
                                    plugin_cactireport_definitions.report_name,
                                    plugin_cactireport_definitions.data_template_id,
                                    plugin_cactireport_definitions.graph_template_id,
                                    plugin_cactireport_definitions.rra_id,
                                    plugin_cactireport_definitions.include_data_from_start_hour,
                                    plugin_cactireport_definitions.include_data_from_end_hour,
                                    plugin_cactireport_definitions.include_data_from_start_day,
                                    plugin_cactireport_definitions.include_data_from_end_day,
                                    rra.name     as rra_name,
                                    rra.steps    as rra_steps,
                                    rra.timespan as rra_timespan
                                  FROM
                                    plugin_cactireport_definitions,
                                    rra
                                  WHERE
                                    plugin_cactireport_definitions.rra_id = rra.id
                                  AND
                                    enabled = 'on'");

    my $removeLastValuesSth = $dbh->prepare("DELETE FROM plugin_cactireport_last_values") || die "Could not prepare last values: " . $dbh->errstr;
    $removeLastValuesSth->execute() || die "Could not execute(): " . $dbh->errstr;
}
else
{
    @reportDefinitions = getRows("SELECT
                                    plugin_cactireport_definitions.id AS report_id,
                                    plugin_cactireport_definitions.report_name,
                                    plugin_cactireport_definitions.data_template_id,
                                    plugin_cactireport_definitions.graph_template_id,
                                    plugin_cactireport_definitions.rra_id,
                                    plugin_cactireport_definitions.include_data_from_start_hour,
                                    plugin_cactireport_definitions.include_data_from_end_hour,
                                    plugin_cactireport_definitions.include_data_from_start_day,
                                    plugin_cactireport_definitions.include_data_from_end_day,
                                    rra.name     as rra_name,
                                    rra.steps    as rra_steps,
                                    rra.timespan as rra_timespan
                                  FROM
                                    plugin_cactireport_definitions,
                                    rra
                                  WHERE
                                    plugin_cactireport_definitions.rra_id = rra.id
                                  AND
                                    rra_id = ?
                                  AND
                                    enabled = 'on'", $rra_id);
}

unless (scalar(@reportDefinitions) > 0)
{
    warn("No report definitions found for RRA ID $rra_id");
}
my $startTime = time();
my $runTime;
foreach my $reportDefinition (@reportDefinitions)
{
    $runTime = $startTime;
    if ($go_to_start == 1)
    {
        $runTime = ($runTime - ($runTime % $reportDefinition->{'rra_timespan'}));
    }
    my @graphs = getGraphsToReportOn($reportDefinition);

    # Add the graph values etc. to each item within each graph
    processGraphData($reportDefinition, @graphs);

    # Now we need to insert the report data into the database.
    storeReportData($reportDefinition, @graphs);
}

# Now clean up some of the older graphs

# We use the following methodology:
# If our current RRA is daily, then we remove all daily reports that are older
# then twice the weekly time-span.
# If Weekly, then remove all weekly reports that are twice the monthly timespan,
# Etc. Etc.
my @rras = getRows("SELECT id, name, steps, timespan FROM rra ORDER BY timespan");

my $found_rra = 0;
my $next_rra;
foreach my $rra (@rras)
{
    if ($rra->{'id'} == $rra_id)
    {
        $found_rra = 1;
        next;
    }
    if ($found_rra == 1)
    {
        $next_rra = $rra;
        last;
    }
}

# The yearly RRA has no sucessessor - ie never delete yearly reports.
if (defined ($next_rra))
{
    # $next_rra is now the next RRA - ie if current RRA is daily, next RRA is weekly.
    my $expireReportsOlderThan = $runTime - ($next_rra->{'timespan'} * 2);
    foreach my $reportDefinition(@reportDefinitions)
    {
        my @reports = getRows("SELECT id FROM plugin_cactireports WHERE report_definition_id = ? AND date_processed < ?",
                              $reportDefinition->{'id'}, $expireReportsOlderThan);

        my $expireReportItemsSth = $dbh->prepare("DELETE FROM plugin_cactireport_items WHERE report_id = ?")
            || die "Could not prepare: " . $dbh->errstr;
        my $expireReportSth = $dbh->prepare("DELETE FROM plugin_cactireports WHERE id = ?") || die "Could not prepare: " . $dbh->errstr;
        foreach my $report (@reports)
        {
            $expireReportItemsSth->execute($report->{'id'}) || die "Could not execute: " . $dbh->errstr;
            $expireReportSth->execute($report->{'id'}) || die "Could not execute: " . $dbh->errstr;
        }
    }
}

## All Done.

exit(0);

sub storeReportData
{
    my $reportDefinition = shift;
    my @graphs           = @_;

    if ($rra_id == 0)
    {
        my $insertLastValueSth = $dbh->prepare("INSERT INTO
                                                  plugin_cactireport_last_values
                                               (
                                                   report_definition_id,
                                                   local_graph_id,
                                                   graph_item_id,
                                                   graph_item_last_value
                                               )
                                               VALUES
                                               (
                                                   ?, ?, ?, ?
                                               )") || die "Could not store last values: " . $dbh->errstr;
        foreach my $graph (@graphs)
        {
            foreach my $graphItem (@{$graph->{'graph_items'}})
            {
                next unless ($graphItem->{'graph_type_id'} >= 4 && $graphItem->{'graph_type_id'} <= 8);
                next if (!defined ($graphItem->{'text_format'}) || $graphItem->{'text_format'} eq '');
                $insertLastValueSth->execute($reportDefinition->{'report_id'},
                                             $graph->{'local_graph_id'},
                                             $graphItem->{'local_graph_template_item_id'},
                                             $graphItem->{'graph_item_last_value'})
                    or die "Could not execute: " . $dbh->errstr;
            }
        }
        return;
    }

    my $createReportSth = $dbh->prepare("INSERT INTO
                                           plugin_cactireports
                                        (
                                           report_definition_id,
                                           date_processed
                                        )
                                        VALUES
                                        (
                                           ?, ?
                                        )") || die "Could not create new report: " . $dbh->errstr;
    $createReportSth->execute($reportDefinition->{'report_id'}, $runTime) || die "Could not execute: " . $dbh->errstr;

    my $reportId;
    if ($dbh->can('last_insert_id'))
    {
        $reportId = $dbh->last_insert_id(undef, undef, 'plugin_cactireports', undef) || $dbh->{'mysql_insertid'} || die "Could not get last insert id";
    }
    else
    {
        $reportId = $dbh->{'mysql_insertid'} || die "Could not get last insert id";
    }

    my $insertResultSth = $dbh->prepare("INSERT INTO
                                           plugin_cactireport_items
                                         (
                                              report_id,
                                              local_graph_id,
                                              graph_item_id,
                                              graph_item_min_value,
                                              graph_item_max_value,
                                              graph_item_avg_value
                                         )
                                         VALUES (?, ?, ?, ?, ?, ?)") || die "Prepare Failed: " . $dbh->errstr;

    foreach my $graph (@graphs)
    {
        foreach my $graphItem (@{$graph->{'graph_items'}})
        {
            next unless ($graphItem->{'graph_type_id'} >= 4 && $graphItem->{'graph_type_id'} <= 8);
            next if (!defined ($graphItem->{'text_format'}) || $graphItem->{'text_format'} eq '');
            $insertResultSth->execute($reportId,
                                      $graph->{'local_graph_id'},
                                      $graphItem->{'local_graph_template_item_id'},
                                      $graphItem->{'graph_item_min_value'} || 0.0,
                                      $graphItem->{'graph_item_max_value'} || 0.0,
                                      $graphItem->{'graph_item_avg_value'} || 0.0)
                or die "Could not execute: " . $dbh->errstr;
        }
    }
}

sub processGraphData
{
    my $reportDefinition = shift;
    my @graphs     = @_;

    my $startTime = localtime($runTime);

    my $path_rra = getRows("SELECT value FROM settings WHERE name = ?", 'path_webroot')->{'value'} . "/rra";
    my @reportData;
    foreach my $graph (@graphs)
    {
        next unless (defined ($graph) && defined ($graph->{'graph_items'}) && scalar(@{$graph->{'graph_items'}}) >= 1);
        # An item can have multiple data sources. (Such as Traffic In and Traffic Out)
        my $def_internal_name  = 'a'; # Is incremented to procude 'a', 'b', 'c' etc.
        my $cdef_internal_name = 'a';
        foreach my $graph_item (@{$graph->{'graph_items'}})
        {
            if (defined ($graph_item->{'data_source_path'}) && defined ($graph_item->{'data_source_name'}))
            {

                $graph_item->{'data_source_path'} =~ s/\<path_rra\>/$path_rra/g;

                if ($graph_item->{'graph_type_id'} >= 4 && $graph_item->{'graph_type_id'} <= 8) # graph_type_id IN (4,5,6,7,8)
                {
                    $graph_item->{'def_internal_name'} = $def_internal_name;
                    $def_internal_name++;
                    $graph_item->{'def'} = sprintf("DEF:%s=%s:%s:%s",
                                                   $graph_item->{'def_internal_name'},
                                                   $graph_item->{'data_source_path'},
                                                   $graph_item->{'data_source_name'},
                                                   $consolidation_functions->{$graph_item->{'consolidation_function_id'}});
                }
            }
        }
        # This has to be done in its own loop becasue all internal_graph_item_name's have to be assigned
        foreach my $graph_item (@{$graph->{'graph_items'}})
        {
            if (defined ($graph_item->{'cdef_id'}) && $graph_item->{'cdef_id'} != 0)
            {
                # Sort out some of the data source particulars...
                # We're going to be exporting the CDEF
                my $cdef;
                $graph_item->{'cdef_internal_name'} = $cdef_internal_name;
                $cdef_internal_name++;

                $cdef = get_cdef_by_id($graph_item->{'cdef_id'});

                if ($cdef =~ /CURRENT_DATA_SOURCE/)
                {
                    my $def_internal_name = $graph_item->{'def_internal_name'} || next;;
                    $cdef =~ s/CURRENT_DATA_SOURCE/$def_internal_name/g;
                }

                if ($cdef =~ /ALL_DATA_SOURCES_(NO)?DUPS/)
                {
                    my @cdef_totals;
                    my $numOtherDataSources = 0;
                    foreach my $otherGraphItem (@{$graph->{'graph_items'}})
                    {
                        next if ($otherGraphItem->{'def_internal_name'} eq $graph_item->{'def_internal_name'} || !defined($otherGraphItem->{'def_internal_name'}));
                        next unless ($otherGraphItem->{'graph_type_id'} >= 4 && $otherGraphItem->{'graph_type_id'} <= 8); # graph_type_id IN (4,5,6,7,8)
                        push @cdef_totals, "TIME",
                                           $runTime - ($otherGraphItem->{'rrd_step'} * $reportDefinition->{'rra_steps'}),
                                           "GT",
                                           $otherGraphItem->{'def_internal_name'},
                                           $otherGraphItem->{'def_internal_name'},
                                           "UN",
                                           "0",
                                           $otherGraphItem->{'def_internal_name'},
                                           "IF",
                                           "IF";
                        $numOtherDataSources++;
                    }
                    my $allgraph_items_cdef = join(',', @cdef_totals) . ',+' x ($numOtherDataSources - 1);
                    $cdef =~ s/ALL_DATA_SOURCES_NODUPS/$allgraph_items_cdef/g;
                }
                if ($cdef =~ /SIMILAR_DATA_SOURCES_(NO)?DUPS/)
                {
                    my @cdef_totals;
                    my $numOtherDataSources = 0;

                    foreach my $otherGraphItem (@{$graph->{'graph_items'}})
                    {
                        next if ($otherGraphItem->{'def_internal_name'} eq $graph_item->{'def_internal_name'} || !defined($otherGraphItem->{'def_internal_name'}));
                        next unless ($otherGraphItem->{'graph_type_id'} >= 4 && $otherGraphItem->{'graph_type_id'} <= 8); # graph_type_id IN (4,5,6,7,8)
                        next unless ($graph_item->{'data_source_name'} eq $otherGraphItem->{'data_source_name'});
                        push @cdef_totals, "TIME",
                                           $runTime - ($otherGraphItem->{'rrd_step'} * $reportDefinition->{'rra_steps'}),
                                           "GT",
                                           $otherGraphItem->{'def_internal_name'},
                                           $otherGraphItem->{'def_internal_name'},
                                           "UN",
                                           "0",
                                           $otherGraphItem->{'def_internal_name'},
                                           "IF",
                                           "IF";
                        $numOtherDataSources++;
                    }
                    my $allgraph_items_cdef = join(',', @cdef_totals) . ',+' x ($numOtherDataSources - 1);
                    $cdef =~ s/SIMILAR_DATA_SOURCES_NODUPS/$allgraph_items_cdef/g;
                }
                if ($cdef && $graph_item->{'cdef_internal_name'})
                {
                    $graph_item->{'cdef'} = sprintf("CDEF:cdef%s=%s", $graph_item->{'cdef_internal_name'}, $cdef);
                }
                if ($graph_item->{'graph_type_id'} >= 4 && $graph_item->{'graph_type_id'} <= 8) # graph_type_id IN (4,5,6,7,8)
                {
                    $graph_item->{'export_name'} = sprintf("XPORT:cdef%s:%s",
                                                            $graph_item->{'cdef_internal_name'},
                                                            $graph_item->{'text_format'});
                }
            }
            else
            {
                if ($graph_item->{'graph_type_id'} >= 4 && $graph_item->{'graph_type_id'} <= 8) # graph_type_id IN (4,5,6,7,8)
                {
                    $graph_item->{'export_name'} = sprintf("XPORT:%s:%s",
                                                            $graph_item->{'def_internal_name'},
                                                            $graph_item->{'text_format'});
                }
            }
        }
        getValuesFromRRDs($reportDefinition, $graph);
    }
}

sub getValuesFromRRDs
{
    my $reportDefinition = shift;
    my $graph            = shift;

    my @defs;
    my @cdefs;
    my @exports;
    my @exportedGraphItems; # Used to keep track of which data sources are exported.
    foreach my $graph_item (@{$graph->{'graph_items'}})
    {
        # def, cdef, export_name
        push @defs,    $graph_item->{'def'}         if ($graph_item->{'def'});
        push @cdefs,   $graph_item->{'cdef'}        if ($graph_item->{'cdef'});
        if ($graph_item->{'export_name'})
        {
            $graph_item->{'graph_item_max_value'} = 0;
            $graph_item->{'graph_item_avg_value'} = 0;
            $graph_item->{'graph_item_last_value'} = 0;
            push @exports, $graph_item->{'export_name'};
            push @exportedGraphItems, $graph_item;
        }
    }

    my @params;

    if ($rra_id == 0) # Allow for getting the current/last value
    {
        @params = ("-m",      $graph->{'width'},
                   "--start", $runTime - 300,
                   "--end",   $runTime,
                   @defs,
                   @cdefs,
                   @exports);
    }
    else
    {
        @params = ("-m",       $graph->{'width'},
                   "--start", $runTime - $reportDefinition->{'rra_timespan'},
                   "--end",   $runTime,
                   @defs,
                   @cdefs,
                   @exports);
    }

    #warn "RRD Export Failed: \n" . Dumper({PARAMS => \@params, GRAPH => $graph});

    # Fetch the data from the RRD...
    my ($start, $end, $step, $col_cnt, $legend, $data) =
      RRDs::xport(@params);
    my $err = RRDs::error;
    if ($err)
    {
        warn "RRD Export Failed: $err\n" . Dumper({PARAMS => \@params, GRAPH => $graph});
        return undef;
    }

    my $numRowsProcessed = 0;
    # Process the data from the RRD...
    foreach my $row (@{$data})
    {
        my $time = localtime($start);
        my $wday = $time->wday;
        $wday = 7 unless ($wday); # Make sunday be 7 instead of 0 (ie Monday is 1)
        $wday--;                  # Then make it zero based so that monday is 0, tuesday is 1...
        if ($wday       >= $reportDefinition->{'include_data_from_start_day'} &&
            $wday       <= $reportDefinition->{'include_data_from_end_day'} &&
            $time->hour >= $reportDefinition->{'include_data_from_start_hour'} &&
            $time->hour <= $reportDefinition->{'include_data_from_end_hour'})
        {
            for (my $i = 0; $i < scalar(@{$row}); $i++)
            {
                # Running total, used for average
                $exportedGraphItems[$i]->{'graph_item_avg_value'} += $row->[$i];
                # Keep a min-value
                if (!defined ($exportedGraphItems[$i]->{'graph_item_min_value'}) || $row->[$i] < $exportedGraphItems[$i]->{'graph_item_min_value'})
                {
                    $exportedGraphItems[$i]->{'graph_item_min_value'} = $row->[$i] if (defined ($row->[$i]));
                }
                # Keep a max value
                $exportedGraphItems[$i]->{'graph_item_max_value'} = $row->[$i] if (defined ($row->[$i]) && $row->[$i] > $exportedGraphItems[$i]->{'graph_item_max_value'});

                # Keep the last value
                if ((!defined ($exportedGraphItems[$i]->{'graph_item_last_value'}) && defined ($row->[$i])) || defined ($row->[$i]))
                {
                    $exportedGraphItems[$i]->{'graph_item_last_value'} = $row->[$i];
                }
            }
            $numRowsProcessed++;
        }
        else
        {
            for (my $i = 0; $i < scalar(@{$row}); $i++)
            {
                if ((!defined ($exportedGraphItems[$i]->{'graph_item_last_value'}) && defined ($row->[$i])) || defined ($row->[$i]))
                {
                    $exportedGraphItems[$i]->{'graph_item_last_value'} = $row->[$i];
                }
            }
        }
        $start += $step;
    }
    foreach my $graph_item (@exportedGraphItems)
    {
        # Make the total into an average
        $graph_item->{'graph_item_avg_value'} /= $numRowsProcessed;
    }
    return @exportedGraphItems;
}

sub get_cdef_by_id
{
    my $id = shift;

    use vars qw($cdef_cache);

    $cdef_cache ||= {};

    if (exists($cdef_cache->{$id}))
    {
        return $cdef_cache->{$id};
    }

    my @cdefs = getRows("SELECT type, value FROM cdef_items WHERE cdef_id = ? ORDER BY sequence", $id);

    my @cdefItems;
    foreach my $cdef (@cdefs)
    {
        if (!defined ($cdef_item_types->{$cdef->{'type'}}))
        {
            push @cdefItems, $cdef->{'value'};
        }
        elsif (ref($cdef_item_types->{$cdef->{'type'}}) eq 'HASH')
        {
            push @cdefItems, $cdef_item_types->{$cdef->{'type'}}->{$cdef->{'value'}};
        }
        elsif (ref($cdef_item_types->{$cdef->{'type'}}) eq 'CODE')
        {
            push @cdefItems, $cdef_item_types->{$cdef->{'type'}}->($cdef->{'value'});
        }
    }
    $cdef_cache->{$id} = join(',', @cdefItems);
    return $cdef_cache->{$id};
}

sub getGraphsToReportOn
{
    my $reportDefinition = shift;

    return undef unless (defined ($reportDefinition->{'data_template_id'}) || defined($reportDefinition->{'graph_template_id'}));

    my ($data_template_query, $graph_template_query) = ('1 = 1', '1 = 1');

    my @args;
    if ($reportDefinition->{'data_template_id'})
    {
        $data_template_query = 'data_template.id = ?';
        push @args, $reportDefinition->{'data_template_id'};
    }
    if ($reportDefinition->{'graph_template_id'})
    {
        $graph_template_query = 'graph_templates_graph.graph_template_id = ?';
        push @args, $reportDefinition->{'graph_template_id'};
    }

    # Get the graphs that will be part of this report
    my @graphs = getRows("SELECT DISTINCT
                            graph_templates_graph.local_graph_id,
                            graph_templates_graph.title_cache AS title,
                            graph_templates_graph.graph_template_id,
                            graph_templates_graph.vertical_label,
                            graph_templates_graph.width
                          FROM
                            data_template,
                            snmp_query_graph_rrd,
                            graph_templates_graph
                          WHERE
                            $data_template_query
                          AND
                            $graph_template_query
                          AND
                            data_template.id = snmp_query_graph_rrd.data_template_id",
                      @args);

    # Things that are plotted on the graph (graph_type_id)
    #<option value="1">COMMENT</option>
    #<option value="2">HRULE</option>
    #<option value="3">VRULE</option>
    #<option value="4">LINE1</option>  --*
    #<option value="5">LINE2</option>    * \
    #<option value="6">LINE3</option>    *  - - These are the only ones we are interested in
    #<option value="7">AREA</option>     * /
    #<option value="8">STACK</option>  --*
    #<option value="9">GPRINT</option>
    #<option value="10">LEGEND</option>

    my $data_template_id_query = '1 = ?';
    my $data_template_id_value = 1;
    if ($reportDefinition->{'data_template_id'})
    {
        $data_template_id_query = 'data_template_rrd.data_template_id = ?';
        $data_template_id_value = $reportDefinition->{'data_template_id'};
        warn("Using Data Template ID: " . $data_template_id_value);
    }


    foreach my $graph (@graphs)
    {
        next unless ($graph->{'local_graph_id'});
        my @graphItems = getRows("SELECT
                                    local_graph_template_item_id,
                                    text_format,
                                    cdef_id,
                                    graph_template_id,
                                    task_item_id,
                                    consolidation_function_id,
                                    graph_type_id
                                  FROM
                                    graph_templates_item
                                  WHERE
                                     consolidation_function_id IN (1,2,3,4)
                                   AND
                                     graph_template_id = ?
                                   AND
                                     local_graph_id = ?
                                   ORDER BY
                                      sequence",
                                  $graph->{'graph_template_id'},
                                  $graph->{'local_graph_id'});
        foreach my $graphItem (@graphItems)
        {
            if ($graphItem->{'task_item_id'})
            {
                my $dataSourceDetails = getRows("SELECT
                                                   data_source_name,
                                                   id,
                                                   local_data_id AS data_source_id,
                                                   data_template_id
                                                 FROM
                                                   data_template_rrd
                                                 WHERE
                                                   id = ?",
                                                 $graphItem->{'task_item_id'});

                $graphItem->{'data_source_name'} = $dataSourceDetails->{'data_source_name'};
                $graphItem->{'data_source_id'}   = $dataSourceDetails->{'data_source_id'};

                my $rrdDetails = getRows("SELECT
                                            data_source_path,
                                            rrd_step,
                                            id AS data_template_id
                                          FROM
                                            data_template_data
                                          WHERE
                                            data_template_data.local_data_id = ?
                                          AND
                                            data_template_id = ?",
                                          $dataSourceDetails->{'data_source_id'},
                                          $dataSourceDetails->{'data_template_id'});

                $graphItem->{'data_source_path'} = $rrdDetails->{'data_source_path'};
                $graphItem->{'rrd_step'}         = $rrdDetails->{'rrd_step'};
                $graphItem->{'data_template_id'} = $rrdDetails->{'data_template_id'};
            }
        }
        $graph->{'graph_items'} = \@graphItems;
        #cluck(Dumper(\@graphItems));
    }
    return @graphs;
}

sub getRows
{
    my $sql    = shift;
    my @params = @_;
    my $sth = $dbh->prepare($sql) || croak "Could not prepare $sql: " . $dbh->errstr;
    $sth->execute(@params) || croak "Could not execute $sql: " . $dbh->errstr;

    my @rows;
    while (my $row = $sth->fetchrow_hashref)
    {
        push @rows, $row;
    }
    if (wantarray)
    {
        return @rows;
    }
    if (scalar(@rows) > 1)
    {
        cluck("$sql returned " . scalar(@rows) . " - Expected 1");
    }
    return $rows[0];
}
